/*
 * Copyright (c) 2005-2010 KOM – Multimedia Communications Lab
 *
 * This file is part of PeerfactSim.KOM.
 * 
 * PeerfactSim.KOM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * 
 * PeerfactSim.KOM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with PeerfactSim.KOM.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.routing;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.PriorityQueue;
import java.util.Queue;
import java.util.concurrent.atomic.DoubleAdder;

import de.tudarmstadt.maki.simonstrator.api.component.vehicular.information.RoadInformation;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.RoadNetwork;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.RoadNetworkEdge;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.RoadNetworkRoute;

public class DijkstraAlgorithm implements RoutingAlgorithm {
	private final double MAX_DEPTH_RATIO = 1.2;

	@Override
	public RoadNetworkRoute findRoute(RoadNetwork pNetwork, RoadNetworkEdge pCurrentPosition, RoadNetworkEdge pDestination,
			List<RoadNetworkRoute> pKnownRoutes) {
		return findRoute(pNetwork, pCurrentPosition, pDestination, pKnownRoutes, Collections.emptyList());
	}

	@Override
	public RoadNetworkRoute findRoute(RoadNetwork pNetwork, RoadNetworkEdge pCurrentPosition,
			RoadNetworkEdge pDestination) {
		return findRoute(pNetwork, pCurrentPosition, pDestination, Collections.emptyList(), Collections.emptyList());
	}

	@Override
	public RoadNetworkRoute findRoute(RoadNetwork pNetwork, RoadNetworkEdge pCurrentPosition, RoadNetworkEdge pDestination,
			List<RoadNetworkRoute> pKnownRoutes, List<RoadNetworkEdge> pEdgesToAvoid) {
		return findRoute(pNetwork, pCurrentPosition, pDestination, pKnownRoutes, pEdgesToAvoid,
				Collections.emptyList());
	}

	@Override
	public RoadNetworkRoute findRoute(RoadNetwork pNetwork, RoadNetworkEdge pCurrentPosition,
			RoadNetworkEdge pDestination, List<RoadNetworkRoute> pKnownRoutes, List<RoadNetworkEdge> pEdgesToAvoid,
			List<RoadInformation> pInfo) {
		if (pCurrentPosition.equals(pDestination)) {
			List<RoadNetworkEdge> segments = new ArrayList<>();
			segments.add(pCurrentPosition);
			return new RoadNetworkRoute(segments);
		}

		Queue<PathInformation> buffer = new PriorityQueue<>(new PathInformationComparator());
		PathInformation rootPathInformation = new PathInformation(pCurrentPosition, null, 0, 0);
		buffer.add(rootPathInformation);

		double currentPathLength = Double.MAX_VALUE;
		RoadNetworkRoute currentShortestRoute = null;

		double maxDepth = 50;
		for (RoadNetworkRoute roadNetworkRoute : pKnownRoutes) {
			if (roadNetworkRoute.getRoute().size() * MAX_DEPTH_RATIO > maxDepth) {
				maxDepth = roadNetworkRoute.getRoute().size() * MAX_DEPTH_RATIO;
			}
		}

		Map<RoadNetworkEdge, DoubleAdder> visited = new HashMap<>();

		outer: while (buffer.size() > 0) {
			PathInformation pathInformation = buffer.poll();

			if (pathInformation.getHops() >= maxDepth) {
				continue;
			}

			if (pathInformation.getLength() >= currentPathLength) {
				continue;
			}

			RoadNetworkEdge edge = pathInformation.getEdge();
			if (visited.containsKey(edge) && visited.get(edge).intValue() > pKnownRoutes.size()) {
				continue;
			}
			if (!visited.containsKey(edge)) {
				visited.put(edge, new DoubleAdder());
			}
			visited.get(edge).add(1);

			if (pEdgesToAvoid.contains(edge)) {
				continue;
			}

			List<RoadNetworkEdge> accessibleEdges = edge.getAccessibleEdges();

			List<RoadNetworkEdge> currentRouteList = new ArrayList<>();
			extractRoute(currentRouteList, pathInformation);
			double edgeLength = edge.calculateStandardEdgeCosts();
			edgeLength += calculateAdditionalEdgeCosts(edge, pInfo,
					pathInformation.getLength() - pCurrentPosition.calculateStandardEdgeCosts(), currentRouteList);
			for (RoadNetworkEdge roadNetworkEdge : accessibleEdges) {
				if (roadNetworkEdge.equals(pDestination)) {
					PathInformation childPathInformation = new PathInformation(roadNetworkEdge, pathInformation,
							pathInformation.getHops(), pathInformation.getLength());
					List<RoadNetworkEdge> routeList = new ArrayList<>();

					extractRoute(routeList, childPathInformation);

					RoadNetworkRoute route = new RoadNetworkRoute(routeList);

					for (RoadNetworkEdge edgeToAvoid : pEdgesToAvoid) {
						if (route.containsEdge(edgeToAvoid)) {
							continue outer;
						}
					}

					if (!pKnownRoutes.contains(route)) {
						if (currentPathLength > childPathInformation.getLength()) {
							currentShortestRoute = route;
							currentPathLength = childPathInformation.getLength();
						}
					}
				} else {
					PathInformation childPathInformation = new PathInformation(roadNetworkEdge, pathInformation,
							pathInformation.getHops() + 1, pathInformation.getLength() + edgeLength);
					buffer.add(childPathInformation);
				}
			}
		}
		return currentShortestRoute;
	}

	protected double calculateAdditionalEdgeCosts(RoadNetworkEdge pEdge, List<RoadInformation> pInfo,
			double pCurrentPathLength, List<RoadNetworkEdge> pCurrentRouteList) {
		if (pInfo != null) {
			for (RoadInformation roadInformation : pInfo) {
				if (roadInformation.getEdge().equals(pEdge)) {
					double additionalCostIfKnown = pEdge.calculateAdditionalCostIfKnown(roadInformation.getValue());
					return additionalCostIfKnown;
				}
			}
		}
		return 0d;
	}

	private void extractRoute(List<RoadNetworkEdge> pEdges, PathInformation pPathInformation) {
		if (pPathInformation.getPreviousEdge() != null) {
			extractRoute(pEdges, pPathInformation.getPreviousEdge());
		}
		pEdges.add(pPathInformation.getEdge());
	}

	private class PathInformationComparator implements Comparator<PathInformation> {

		@Override
		public int compare(PathInformation pO1, PathInformation pO2) {
			return Double.compare(pO1.getLength(), pO2.getLength());
		}

	}

	private class PathInformation {
		private final int _hops;
		private final double _length;
		private final RoadNetworkEdge _edge;
		private final PathInformation _previousEdge;

		public PathInformation(RoadNetworkEdge pEdge,
				PathInformation pPreviousEdge, int pHops, double pLenght) {
			_edge = pEdge;
			_previousEdge = pPreviousEdge;
			_hops = pHops;
			_length = pLenght;
		}

		/**
		 * @return
		 */
		public double getLength() {
			return _length;
		}

		public int getHops() {
			return _hops;
		}

		public RoadNetworkEdge getEdge() {
			return _edge;
		}

		public PathInformation getPreviousEdge() {
			return _previousEdge;
		}

		@Override
		public boolean equals(Object pObj) {
			if (pObj instanceof PathInformation) {
				return ((PathInformation) pObj).getEdge().equals(_edge);
			}
			return false;
		}
	}
}
